local SCRIPT_VERSION = 0

local fetch
do
  local curl = optional.curl
  assert(curl.initialize(os.type == 'windows' and 'libcurl.dll' or 'libcurl.so'))
  --[[
  local opts = (function()
    local f = assert(io.open'/usr/include/curl/curl.h')
    local CURLOPT = 'CURLOPT(CURLOPT_' * (lpeg.R('AZ', '09', '__')^1 / string.lower) * ', CURLOPTTYPE_' * (lpeg.P'LONG' + 'OFF_T' + 'BLOB' + 'STRINGPOINT' + 'SLISTPOINT' + 'VALUES') * ', ' * (lpeg.R'09'^1/tonumber) * ')'
    local opts = assert(lpeg.Cf(lpeg.Ct'' * (lpeg.Cg(CURLOPT) + 1)^0, rawset):match(f:read'a'))
    f:close()
    return opts
  end)()
  ]]
  local opts = {
    url = 2,
    failonerror = 45,
    followlocation = 52,
  }
  function fetch(args)
    local indexed = {
      [opts.followlocation] = true,
      [opts.failonerror] = true,
    }
    for k, v in pairs(args) do
      indexed[opts[k]] = v
    end
    return curl.fetch(indexed)
  end
end

local base_url = 'https://lmltx.typesetting.eu/'

local platform_url
do
  local manifest = fetch{url = base_url .. 'manifest'}
  if not manifest then
    error'Unable to reach download server'
  end
  manifest = assert(load(manifest, 'manifest', 't', {}))
  manifest = assert(manifest())

  if manifest.version ~= SCRIPT_VERSION then
    error'Installer outdated. Please download a new version.'
  end

  local os_type, os_name, os_machine = os.type, os.name, os.uname().machine
  local this_platform = manifest[os_type]
  this_platform = this_platform and this_platform[os_name]
  this_platform = this_platform and this_platform[os_machine]
  
  if not this_platform then
    error'This platform is currently not supported'
  end
  platform_url = this_platform .. '/'
end

local platform_manifest = assert(fetch{url = platform_url .. 'manifest'})
platform_manifest = assert(load(platform_manifest, 'manifest', 't', {}))
platform_manifest = assert(platform_manifest())

local function get_line(cmd)
  local stream = io.popen(cmd)
  if not stream then return end
  local result = stream:read'l'
  local after = stream:read'l'
  if after then error'More output than expected' end
  if not stream:close() then return end
  return result
end

local binary_location, luametalatex_location
do
  binary_location = get_line 'kpsewhich -var-value SELFAUTOLOC'
  if not binary_location then
    error'Unable to find TeX Live binary directory'
  end
  if arg[1] and lfs.isfile(arg[1] .. '/luametalatex.lua') then
    luametalatex_location = arg[1]
  else
    luametalatex_location = get_line 'kpsewhich luametalatex-init.lua'
    if not luametalatex_location then
      error'Unable to find LuaMetaLaTeX install location. Did you run l3build install?'
    end
    luametalatex_location = assert(luametalatex_location:match'^(.*)/luametalatex%-init%.lua$')
  end
end

local function update_file(entry, basepath)
  local up_to_date
  local path = basepath .. '/' .. entry.filename
  local f = io.open(path, 'rb')
  if f then
    local hash = sha2.hex256(f:read'a')
    if hash == entry.sha256 then
      up_to_date = true
    end
    f:close()
  end
  if not up_to_date then
    local content = assert(fetch{url = platform_url .. entry.filename})
    if sha2.hex256(content) ~= entry.sha256 then
      error'Corrupted download'
    end
    f = assert(io.open(path, 'wb'))
    f:write(content)
    f:close()
    if entry.executable then
      assert(lfs.setexecutable(path))
    end
  end
end

local function try_link(source, target)
  if lfs.isfile(target) then
    assert(os.remove(target))
  end
  if not lfs.symlink(source, target) then
    local output = assert(io.open(target, 'wb'))
    local input = assert(io.open(source, 'rb'))
    output:write(input:read'a')
    input:close()
    output:close()
  end
end

-- Prerequisite: base is an existing directory
local function mkdirp(base, name, ...)
  if not name then return base end
  base = base .. '/' .. name
  return (lfs.isdir(base) or lfs.mkdir(base)) and mkdirp(base, ...)
end

local exe_ext = os.type == 'windows' and '.exe' or ''

update_file(platform_manifest.engine, binary_location)

local library_location = mkdirp(binary_location, 'lib', 'luametatex', 'lua')

update_file(platform_manifest.library, library_location)

try_link(binary_location .. '/luametatex' .. exe_ext, binary_location .. '/luametaplain' .. exe_ext)
try_link(binary_location .. '/luametatex' .. exe_ext, binary_location .. '/luametalatex' .. exe_ext)
try_link(binary_location .. '/luametatex' .. exe_ext, binary_location .. '/luametalatex-dev' .. exe_ext)
try_link(luametalatex_location .. '/luametalatex.lua', binary_location .. '/luametaplain.lua')
try_link(luametalatex_location .. '/luametalatex.lua', binary_location .. '/luametalatex.lua')
try_link(luametalatex_location .. '/luametalatex.lua', binary_location .. '/luametalatex-dev.lua')

local formats = {}
do
  local f = assert(io.popen('fmtutil -sys -listcfg -quiet', 'r'))
  assert(f:read'l' == "List of all formats:")
  while true do
    local line = f:read'l'
    if not line then break end
    local format, engine, state = line:match'^(.*) %(engine=(.*)%) (.*)$'
    line = f:read'l'
    local hyph, args = line:match'^  hyphen=(.*), args=(.*)$'
    line = f:read'l'
    local origin = line:match'^  origin=(.*)$'
    if engine == 'luametatex' then
      formats[format] = state == 'enabled'
    end
  end
  assert(f:close())
end
if next(formats) then
  local format_found
  for _, fmt in ipairs{'luametalatex', 'luametalatex-dev', 'luametaplain'} do
    if formats[fmt] == nil then
      print(string.format('Warning: Format information has not been found for %s.', fmt))
    elseif formats[fmt] then
      format_found = true
    end
  end
  if not format_found then
    print'Warning: All LuaMetaTeX formats inactive. You can still install, but will not be able to actually use LuaMetaLaTeX like this.'
  end
else
  local config_dir = assert(get_line 'kpsewhich -var-value=TEXMFSYSCONFIG') -- /web2c/fmtutil.cnf"')
  if not lfs.isdir(config_dir) then
    assert(lfs.mkdir(config_dir))
  end
  config_dir = config_dir .. '/web2c'
  if not lfs.isdir(config_dir) then
    assert(lfs.mkdir(config_dir))
  end
  local f = assert(io.open(config_dir .. '/fmtutil.cnf', 'a'))
  local driver = os.type == 'windows' and luametalatex_location .. '/luametalatex.lua' or '$(kpsewhich luametalatex.lua)'
  f:write(string.format(
    '\z
      luametaplain luametatex language.dat,language.dat.lua --lua="%s" luametaplain.ini\n\z
      luametalatex luametatex language.dat,language.dat.lua --lua="%s" luametalatex.ini\n\z
      luametalatex-dev luametatex language.dat,language.dat.lua --lua="%s" luametalatex.ini\n\z
    ',
    driver,
    driver,
    driver
  ))
  f:close()
  f = assert(io.open(config_dir .. '/texmf.cnf', 'a'))
  f:write(string.format(
    '\z
      TEXINPUTS.luametaplain    = $TEXMFDOTDIR;$TEXMF/tex/{luametaplain,luametatex,luatex,plain,generic,}//\n\z
      TEXINPUTS.luametalatex    = $TEXMFDOTDIR;$TEXMF/tex/{luametalatex,lualatex,latex,luametatex,luatex,generic,}//\n\z
      TEXINPUTS.luametalatex-dev= $TEXMFDOTDIR;$TEXMF/tex/{luametalatex,latex-dev,lualatex,latex,luametatex,luatex,generic,}//\n\z
      \n\z
      LUAINPUTS.luametaplain    = $TEXMFDOTDIR;$TEXMF/scripts/{$progname,$engine,}/{lua,}//;$TEXMF/tex/{luametaplain,luametatex,luatex,plain,generic,}//\n\z
      LUAINPUTS.luametalatex    = $TEXMFDOTDIR;$TEXMF/scripts/{$progname,$engine,}/{lua,}//;$TEXMF/tex/{luametalatex,lualatex,latex,luametatex,luatex,generic,}//\n\z
      LUAINPUTS.luametalatex-dev= $TEXMFDOTDIR;$TEXMF/scripts/{$progname,$engine,}/{lua,}//;$TEXMF/tex/{luametalatex,latex-dev,lualatex,latex,luametatex,luatex,generic,}//\n\z
    '
  ))
  f:close()
  assert(os.execute'mktexlsr')
end
